@@ -33,6 +33,8 @@
 #include <grub/efi/api.h>
 #endif
 
+#pragma GCC diagnostic ignored "-Wcast-align"
+
 GRUB_MOD_LICENSE ("GPLv3+");
 
 static const struct grub_arg_option options[] = {
@@ -41,8 +43,8 @@ static const struct grub_arg_option options[] = {
    0, ARG_TYPE_STRING},
   {"load-only", 'n', 0,
    N_("Load only tables specified by comma-separated list."), 0, ARG_TYPE_STRING},
-  {"v1", '1', 0, N_("Expose v1 tables."), 0, ARG_TYPE_NONE},
-  {"v2", '2', 0, N_("Expose v2 and v3 tables."), 0, ARG_TYPE_NONE},
+  {"v1", '1', 0, N_("Export version 1 tables to the OS."), 0, ARG_TYPE_NONE},
+  {"v2", '2', 0, N_("Export version 2 and version 3 tables to the OS."), 0, ARG_TYPE_NONE},
   {"oemid", 'o', 0, N_("Set OEMID of RSDP, XSDT and RSDT."), 0, ARG_TYPE_STRING},
   {"oemtable", 't', 0,
    N_("Set OEMTABLE ID of RSDP, XSDT and RSDT."), 0, ARG_TYPE_STRING},
@@ -52,24 +54,13 @@ static const struct grub_arg_option options[] = {
    N_("Set creator field of RSDP, XSDT and RSDT."), 0, ARG_TYPE_STRING},
   {"oemtablecreatorrev", 'd', 0,
    N_("Set creator revision of RSDP, XSDT and RSDT."), 0, ARG_TYPE_INT},
-  {"no-ebda", 'e', 0, N_("Don't update EBDA. May fix failures or hangs on some."
-   " BIOSes but makes it ineffective with OS not receiving RSDP from GRUB."),
+  /* TRANSLATORS: "hangs" here is a noun, not a verb.  */
+  {"no-ebda", 'e', 0, N_("Don't update EBDA. May fix failures or hangs on some "
+   "BIOSes but makes it ineffective with OS not receiving RSDP from GRUB."),
    0, ARG_TYPE_NONE},
   {0, 0, 0, 0, 0, 0}
 };
 
-/* Simple checksum by summing all bytes. Used by ACPI and SMBIOS. */
-grub_uint8_t
-grub_byte_checksum (void *base, grub_size_t size)
-{
-  grub_uint8_t *ptr;
-  grub_uint8_t ret = 0;
-  for (ptr = (grub_uint8_t *) base; ptr < ((grub_uint8_t *) base) + size;
-       ptr++)
-    ret += *ptr;
-  return ret;
-}
-
 /* rev1 is 1 if ACPIv1 is to be generated, 0 otherwise.
    rev2 contains the revision of ACPIv2+ to generate or 0 if none. */
 static int rev1, rev2;
@@ -128,6 +119,8 @@ grub_acpi_get_rsdpv1 (void)
   return grub_machine_acpi_get_rsdpv1 ();
 }
 
+#if defined (__i386__) || defined (__x86_64__)
+
 static inline int
 iszero (grub_uint8_t *reg, int size)
 {
@@ -138,49 +131,61 @@ iszero (grub_uint8_t *reg, int size)
   return 1;
 }
 
+/* Context for grub_acpi_create_ebda.  */
+struct grub_acpi_create_ebda_ctx {
+  int ebda_len;
+  grub_uint64_t highestlow;
+};
+
+/* Helper for grub_acpi_create_ebda.  */
+static int
+find_hook (grub_uint64_t start, grub_uint64_t size, grub_memory_type_t type,
+	   void *data)
+{
+  struct grub_acpi_create_ebda_ctx *ctx = data;
+  grub_uint64_t end = start + size;
+  if (type != GRUB_MEMORY_AVAILABLE)
+    return 0;
+  if (end > 0x100000)
+    end = 0x100000;
+  if (end > start + ctx->ebda_len
+      && ctx->highestlow < ((end - ctx->ebda_len) & (~0xf)) )
+    ctx->highestlow = (end - ctx->ebda_len) & (~0xf);
+  return 0;
+}
+
 grub_err_t
 grub_acpi_create_ebda (void)
 {
-  int ebda_kb_len;
-  int ebda_len;
+  struct grub_acpi_create_ebda_ctx ctx = {
+    .highestlow = 0
+  };
+  int ebda_kb_len = 0;
   int mmapregion = 0;
   grub_uint8_t *ebda, *v1inebda = 0, *v2inebda = 0;
-  grub_uint64_t highestlow = 0;
   grub_uint8_t *targetebda, *target;
   struct grub_acpi_rsdp_v10 *v1;
   struct grub_acpi_rsdp_v20 *v2;
-  auto int NESTED_FUNC_ATTR find_hook (grub_uint64_t, grub_uint64_t,
-				       grub_uint32_t);
-  int NESTED_FUNC_ATTR find_hook (grub_uint64_t start, grub_uint64_t size,
-				  grub_memory_type_t type)
-  {
-    grub_uint64_t end = start + size;
-    if (type != GRUB_MEMORY_AVAILABLE)
-      return 0;
-    if (end > 0x100000)
-      end = 0x100000;
-    if (end > start + ebda_len
-	&& highestlow < ((end - ebda_len) & (~0xf)) )
-      highestlow = (end - ebda_len) & (~0xf);
-    return 0;
-  }
 
-  ebda = (grub_uint8_t *) UINT_TO_PTR ((*((grub_uint16_t *)0x40e)) << 4);
-  ebda_kb_len = *(grub_uint16_t *) ebda;
-  if (! ebda || ebda_kb_len > 16)
+  ebda = (grub_uint8_t *) (grub_addr_t) ((*((grub_uint16_t *)0x40e)) << 4);
+  grub_dprintf ("acpi", "EBDA @%p\n", ebda);
+  if (ebda)
+    ebda_kb_len = *(grub_uint16_t *) ebda;
+  grub_dprintf ("acpi", "EBDA length 0x%x\n", ebda_kb_len);
+  if (ebda_kb_len > 16)
     ebda_kb_len = 0;
-  ebda_len = (ebda_kb_len + 1) << 10;
+  ctx.ebda_len = (ebda_kb_len + 1) << 10;
 
   /* FIXME: use low-memory mm allocation once it's available. */
-  grub_mmap_iterate (find_hook);
-  targetebda = (grub_uint8_t *) UINT_TO_PTR (highestlow);
+  grub_mmap_iterate (find_hook, &ctx);
+  targetebda = (grub_uint8_t *) (grub_addr_t) ctx.highestlow;
   grub_dprintf ("acpi", "creating ebda @%llx\n",
-		(unsigned long long) highestlow);
-  if (! highestlow)
+		(unsigned long long) ctx.highestlow);
+  if (! ctx.highestlow)
     return grub_error (GRUB_ERR_OUT_OF_MEMORY,
 		       "couldn't find space for the new EBDA");
 
-  mmapregion = grub_mmap_register (PTR_TO_UINT64 (targetebda), ebda_len,
+  mmapregion = grub_mmap_register ((grub_addr_t) targetebda, ctx.ebda_len,
 				   GRUB_MEMORY_RESERVED);
   if (! mmapregion)
     return grub_errno;
@@ -203,7 +208,7 @@ grub_acpi_create_ebda (void)
     {
       grub_dprintf ("acpi", "Scanning EBDA for old rsdpv2\n");
       for (; target < targetebda + 0x400 - v2->length; target += 0x10)
-	if (grub_memcmp (target, "RSD PTR ", 8) == 0
+	if (grub_memcmp (target, GRUB_RSDP_SIGNATURE, GRUB_RSDP_SIGNATURE_SIZE) == 0
 	    && grub_byte_checksum (target,
 				   sizeof (struct grub_acpi_rsdp_v10)) == 0
 	    && ((struct grub_acpi_rsdp_v10 *) target)->revision != 0
@@ -213,7 +218,7 @@ grub_acpi_create_ebda (void)
 	    grub_dprintf ("acpi", "Copying rsdpv2 to %p\n", target);
 	    v2inebda = target;
 	    target += v2->length;
-	    target = (grub_uint8_t *) ((((long) target - 1) | 0xf) + 1);
+	    target = (grub_uint8_t *) ALIGN_UP((grub_addr_t) target, 16);
 	    v2 = 0;
 	    break;
 	  }
@@ -224,7 +229,7 @@ grub_acpi_create_ebda (void)
       grub_dprintf ("acpi", "Scanning EBDA for old rsdpv1\n");
       for (; target < targetebda + 0x400 - sizeof (struct grub_acpi_rsdp_v10);
 	   target += 0x10)
-	if (grub_memcmp (target, "RSD PTR ", 8) == 0
+	if (grub_memcmp (target, GRUB_RSDP_SIGNATURE, GRUB_RSDP_SIGNATURE_SIZE) == 0
 	    && grub_byte_checksum (target,
 				   sizeof (struct grub_acpi_rsdp_v10)) == 0)
 	  {
@@ -232,7 +237,7 @@ grub_acpi_create_ebda (void)
 	    grub_dprintf ("acpi", "Copying rsdpv1 to %p\n", target);
 	    v1inebda = target;
 	    target += sizeof (struct grub_acpi_rsdp_v10);
-	    target = (grub_uint8_t *) ((((long) target - 1) | 0xf) + 1);
+	    target = (grub_uint8_t *) ALIGN_UP((grub_addr_t) target, 16);
 	    v1 = 0;
 	    break;
 	  }
@@ -251,7 +256,7 @@ grub_acpi_create_ebda (void)
 	    grub_memcpy (target, v2, v2->length);
 	    v2inebda = target;
 	    target += v2->length;
-	    target = (grub_uint8_t *) ((((long) target - 1) | 0xf) + 1);
+	    target = (grub_uint8_t *) ALIGN_UP((grub_addr_t) target, 16);
 	    v2 = 0;
 	    break;
 	  }
@@ -268,7 +273,7 @@ grub_acpi_create_ebda (void)
 	    grub_memcpy (target, v1, sizeof (struct grub_acpi_rsdp_v10));
 	    v1inebda = target;
 	    target += sizeof (struct grub_acpi_rsdp_v10);
-	    target = (grub_uint8_t *) ((((long) target - 1) | 0xf) + 1);
+	    target = (grub_uint8_t *) ALIGN_UP((grub_addr_t) target, 16);
 	    v1 = 0;
 	    break;
 	  }
@@ -285,18 +290,19 @@ grub_acpi_create_ebda (void)
   for (target = targetebda;
        target < targetebda + 0x400 - sizeof (struct grub_acpi_rsdp_v10);
        target += 0x10)
-    if (grub_memcmp (target, "RSD PTR ", 8) == 0
+    if (grub_memcmp (target, GRUB_RSDP_SIGNATURE, GRUB_RSDP_SIGNATURE_SIZE) == 0
 	&& grub_byte_checksum (target,
 			       sizeof (struct grub_acpi_rsdp_v10)) == 0
 	&& target != v1inebda && target != v2inebda)
       *target = 0;
 
   grub_dprintf ("acpi", "Switching EBDA\n");
-  (*((grub_uint16_t *) 0x40e)) = ((long)targetebda) >> 4;
+  (*((grub_uint16_t *) 0x40e)) = ((grub_addr_t) targetebda) >> 4;
   grub_dprintf ("acpi", "EBDA switched\n");
 
   return GRUB_ERR_NONE;
 }
+#endif
 
 /* Create tables common to ACPIv1 and ACPIv2+ */
 static void
@@ -328,13 +334,13 @@ setup_common_tables (void)
       if (grub_memcmp (fadt->hdr.signature, GRUB_ACPI_FADT_SIGNATURE,
 		       sizeof (fadt->hdr.signature)) == 0)
 	{
-	  fadt->dsdt_addr = PTR_TO_UINT32 (table_dsdt);
+	  fadt->dsdt_addr = (grub_addr_t) table_dsdt;
 	  fadt->facs_addr = facs_addr;
 
 	  /* Does a revision 2 exist at all? */
 	  if (fadt->hdr.revision >= 3)
 	    {
-	      fadt->dsdt_xaddr = PTR_TO_UINT64 (table_dsdt);
+	      fadt->dsdt_xaddr = (grub_addr_t) table_dsdt;
 	      fadt->facs_xaddr = facs_addr;
 	    }
 
@@ -350,13 +356,13 @@ setup_common_tables (void)
     numoftables++;
 
   rsdt_addr = rsdt = (struct grub_acpi_table_header *) playground_ptr;
-  playground_ptr += sizeof (struct grub_acpi_table_header) + 4 * numoftables;
+  playground_ptr += sizeof (struct grub_acpi_table_header) + sizeof (grub_uint32_t) * numoftables;
 
   rsdt_entry = (grub_uint32_t *) (rsdt + 1);
 
   /* Fill RSDT header. */
   grub_memcpy (&(rsdt->signature), "RSDT", 4);
-  rsdt->length = sizeof (struct grub_acpi_table_header) + 4 * numoftables;
+  rsdt->length = sizeof (struct grub_acpi_table_header) + sizeof (grub_uint32_t) * numoftables;
   rsdt->revision = 1;
   grub_memcpy (&(rsdt->oemid), root_oemid, sizeof (rsdt->oemid));
   grub_memcpy (&(rsdt->oemtable), root_oemtable, sizeof (rsdt->oemtable));
@@ -365,7 +371,7 @@ setup_common_tables (void)
   rsdt->creator_rev = root_creator_rev;
 
   for (cur = acpi_tables; cur; cur = cur->next)
-    *(rsdt_entry++) = PTR_TO_UINT32 (cur->addr);
+    *(rsdt_entry++) = (grub_addr_t) cur->addr;
 
   /* Recompute checksum. */
   rsdt->checksum = 0;
@@ -379,11 +385,11 @@ setv1table (void)
   /* Create RSDP. */
   rsdpv1_new = (struct grub_acpi_rsdp_v10 *) playground_ptr;
   playground_ptr += sizeof (struct grub_acpi_rsdp_v10);
-  grub_memcpy (&(rsdpv1_new->signature), "RSD PTR ",
+  grub_memcpy (&(rsdpv1_new->signature), GRUB_RSDP_SIGNATURE,
 	       sizeof (rsdpv1_new->signature));
   grub_memcpy (&(rsdpv1_new->oemid), root_oemid, sizeof  (rsdpv1_new->oemid));
   rsdpv1_new->revision = 0;
-  rsdpv1_new->rsdt_addr = PTR_TO_UINT32 (rsdt_addr);
+  rsdpv1_new->rsdt_addr = (grub_addr_t) rsdt_addr;
   rsdpv1_new->checksum = 0;
   rsdpv1_new->checksum = 1 + ~grub_byte_checksum (rsdpv1_new,
 						  sizeof (*rsdpv1_new));
@@ -404,13 +410,13 @@ setv2table (void)
 
   /* Create XSDT. */
   xsdt = (struct grub_acpi_table_header *) playground_ptr;
-  playground_ptr += sizeof (struct grub_acpi_table_header) + 8 * numoftables;
+  playground_ptr += sizeof (struct grub_acpi_table_header) + sizeof (grub_uint64_t) * numoftables;
 
   xsdt_entry = (grub_uint64_t *)(xsdt + 1);
   for (cur = acpi_tables; cur; cur = cur->next)
-    *(xsdt_entry++) = PTR_TO_UINT64 (cur->addr);
+    *(xsdt_entry++) = (grub_addr_t) cur->addr;
   grub_memcpy (&(xsdt->signature), "XSDT", 4);
-  xsdt->length = sizeof (struct grub_acpi_table_header) + 8 * numoftables;
+  xsdt->length = sizeof (struct grub_acpi_table_header) + sizeof (grub_uint64_t) * numoftables;
   xsdt->revision = 1;
   grub_memcpy (&(xsdt->oemid), root_oemid, sizeof (xsdt->oemid));
   grub_memcpy (&(xsdt->oemtable), root_oemtable, sizeof (xsdt->oemtable));
@@ -423,17 +429,17 @@ setv2table (void)
   /* Create RSDPv2. */
   rsdpv2_new = (struct grub_acpi_rsdp_v20 *) playground_ptr;
   playground_ptr += sizeof (struct grub_acpi_rsdp_v20);
-  grub_memcpy (&(rsdpv2_new->rsdpv1.signature), "RSD PTR ",
+  grub_memcpy (&(rsdpv2_new->rsdpv1.signature), GRUB_RSDP_SIGNATURE,
 	       sizeof (rsdpv2_new->rsdpv1.signature));
   grub_memcpy (&(rsdpv2_new->rsdpv1.oemid), root_oemid,
 	       sizeof (rsdpv2_new->rsdpv1.oemid));
   rsdpv2_new->rsdpv1.revision = rev2;
-  rsdpv2_new->rsdpv1.rsdt_addr = PTR_TO_UINT32 (rsdt_addr);
+  rsdpv2_new->rsdpv1.rsdt_addr = (grub_addr_t) rsdt_addr;
   rsdpv2_new->rsdpv1.checksum = 0;
   rsdpv2_new->rsdpv1.checksum = 1 + ~grub_byte_checksum
     (&(rsdpv2_new->rsdpv1), sizeof (rsdpv2_new->rsdpv1));
   rsdpv2_new->length = sizeof (*rsdpv2_new);
-  rsdpv2_new->xsdt_addr = PTR_TO_UINT64 (xsdt);
+  rsdpv2_new->xsdt_addr = (grub_addr_t) xsdt;
   rsdpv2_new->checksum = 0;
   rsdpv2_new->checksum = 1 + ~grub_byte_checksum (rsdpv2_new,
 						  rsdpv2_new->length);
@@ -463,7 +469,6 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
   struct grub_arg_list *state = ctxt->state;
   struct grub_acpi_rsdp_v10 *rsdp;
   struct efiemu_acpi_table *cur, *t;
-  grub_err_t err;
   int i, mmapregion;
   int numoftables;
 
@@ -480,6 +485,8 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
   if (! rsdp)
     rsdp = grub_machine_acpi_get_rsdpv1 ();
 
+  grub_dprintf ("acpi", "RSDP @%p\n", rsdp);
+
   if (rsdp)
     {
       grub_uint32_t *entry_ptr;
@@ -506,7 +513,7 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
       /* Set revision variables to replicate the same version as host. */
       rev1 = ! rsdp->revision;
       rev2 = rsdp->revision;
-      rsdt = (struct grub_acpi_table_header *) UINT_TO_PTR (rsdp->rsdt_addr);
+      rsdt = (struct grub_acpi_table_header *) (grub_addr_t) rsdp->rsdt_addr;
       /* Load host tables. */
       for (entry_ptr = (grub_uint32_t *) (rsdt + 1);
 	   entry_ptr < (grub_uint32_t *) (((grub_uint8_t *) rsdt)
@@ -516,7 +523,7 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 	  char signature[5];
 	  struct efiemu_acpi_table *table;
 	  struct grub_acpi_table_header *curtable
-	    = (struct grub_acpi_table_header *) UINT_TO_PTR (*entry_ptr);
+	    = (struct grub_acpi_table_header *) (grub_addr_t) *entry_ptr;
 	  signature[4] = 0;
 	  for (i = 0; i < 4;i++)
 	    signature[i] = grub_tolower (curtable->signature[i]);
@@ -540,7 +547,7 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 
 	      /* Load DSDT if not excluded. */
 	      dsdt = (struct grub_acpi_table_header *)
-		UINT_TO_PTR (fadt->dsdt_addr);
+		(grub_addr_t) fadt->dsdt_addr;
 	      if (dsdt && (! exclude || ! grub_strword (exclude, "dsdt"))
 		  && (! load_only || grub_strword (load_only, "dsdt"))
 		  && dsdt->length >= sizeof (*dsdt))
@@ -552,8 +559,7 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 		      free_tables ();
 		      grub_free (exclude);
 		      grub_free (load_only);
-		      return grub_error (GRUB_ERR_OUT_OF_MEMORY,
-					 "couldn't allocate table");
+		      return grub_errno;
 		    }
 		  grub_memcpy (table_dsdt, dsdt, dsdt->length);
 		}
@@ -579,8 +585,7 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 	      free_tables ();
 	      grub_free (exclude);
 	      grub_free (load_only);
-	      return grub_error (GRUB_ERR_OUT_OF_MEMORY,
-				 "couldn't allocate table structure");
+	      return grub_errno;
 	    }
 	  table->size = curtable->length;
 	  table->addr = grub_malloc (table->size);
@@ -588,8 +593,10 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 	  if (! table->addr)
 	    {
 	      free_tables ();
-	      return grub_error (GRUB_ERR_OUT_OF_MEMORY,
-				 "couldn't allocate table");
+	      grub_free (exclude);
+	      grub_free (load_only);
+	      grub_free (table);
+	      return grub_errno;
 	    }
 	  table->next = acpi_tables;
 	  acpi_tables = table;
@@ -628,11 +635,11 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
       grub_size_t size;
       char *buf;
 
-      file = grub_file_open (args[i]);
+      file = grub_file_open (args[i], GRUB_FILE_TYPE_ACPI_TABLE);
       if (! file)
 	{
 	  free_tables ();
-	  return grub_error (GRUB_ERR_BAD_OS, "couldn't open file %s", args[i]);
+	  return grub_errno;
 	}
 
       size = grub_file_size (file);
@@ -640,7 +647,8 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 	{
 	  grub_file_close (file);
 	  free_tables ();
-	  return grub_error (GRUB_ERR_BAD_OS, "file %s is too small", args[i]);
+	  return grub_error (GRUB_ERR_BAD_OS, N_("premature end of file %s"),
+			     args[i]);
 	}
 
       buf = (char *) grub_malloc (size);
@@ -648,15 +656,17 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 	{
 	  grub_file_close (file);
 	  free_tables ();
-	  return grub_error (GRUB_ERR_OUT_OF_MEMORY,
-			     "couldn't read file %s", args[i]);
+	  return grub_errno;
 	}
 
       if (grub_file_read (file, buf, size) != (int) size)
 	{
 	  grub_file_close (file);
 	  free_tables ();
-	  return grub_error (GRUB_ERR_BAD_OS, "couldn't read file %s", args[i]);
+	  if (!grub_errno)
+	    grub_error (GRUB_ERR_BAD_OS, N_("premature end of file %s"),
+			args[i]);
+	  return grub_errno;
 	}
       grub_file_close (file);
 
@@ -675,8 +685,7 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
 	  if (! table)
 	    {
 	      free_tables ();
-	      return grub_error (GRUB_ERR_OUT_OF_MEMORY,
-				 "couldn't allocate table structure");
+	      return grub_errno;
 	    }
 
 	  table->size = size;
@@ -695,11 +704,11 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
   /* DSDT. */
   playground_size += dsdt_size;
   /* RSDT. */
-  playground_size += sizeof (struct grub_acpi_table_header) + 4 * numoftables;
+  playground_size += sizeof (struct grub_acpi_table_header) + sizeof (grub_uint32_t) * numoftables;
   /* RSDPv1. */
   playground_size += sizeof (struct grub_acpi_rsdp_v10);
   /* XSDT. */
-  playground_size += sizeof (struct grub_acpi_table_header) + 8 * numoftables;
+  playground_size += sizeof (struct grub_acpi_table_header) + sizeof (grub_uint64_t) * numoftables;
   /* RSDPv2. */
   playground_size += sizeof (struct grub_acpi_rsdp_v20);
 
@@ -732,23 +741,30 @@ grub_cmd_acpi (struct grub_extcmd_context *ctxt, int argc, char **args)
     }
   acpi_tables = 0;
 
-  if (! state[9].set && (err = grub_acpi_create_ebda ()))
+#if defined (__i386__) || defined (__x86_64__)
+  if (! state[9].set)
     {
-      rsdpv1_new = 0;
-      rsdpv2_new = 0;
-      grub_mmap_free_and_unregister (mmapregion);
-      return err;
+      grub_err_t err;
+      err = grub_acpi_create_ebda ();
+      if (err)
+	{
+	  rsdpv1_new = 0;
+	  rsdpv2_new = 0;
+	  grub_mmap_free_and_unregister (mmapregion);
+	  return err;
+	}
     }
+#endif
 
 #ifdef GRUB_MACHINE_EFI
   {
     struct grub_efi_guid acpi = GRUB_EFI_ACPI_TABLE_GUID;
     struct grub_efi_guid acpi20 = GRUB_EFI_ACPI_20_TABLE_GUID;
 
-    grub_efi_system_table->boot_services->install_configuration_table
-      (&acpi20, grub_acpi_get_rsdpv2 ());
-    grub_efi_system_table->boot_services->install_configuration_table
-      (&acpi, grub_acpi_get_rsdpv1 ());
+    efi_call_2 (grub_efi_system_table->boot_services->install_configuration_table,
+      &acpi20, grub_acpi_get_rsdpv2 ());
+    efi_call_2 (grub_efi_system_table->boot_services->install_configuration_table,
+      &acpi, grub_acpi_get_rsdpv1 ());
   }
 #endif
 
@@ -761,7 +777,7 @@ GRUB_MOD_INIT(acpi)
 {
   cmd = grub_register_extcmd ("acpi", grub_cmd_acpi, 0,
 			      N_("[-1|-2] [--exclude=TABLE1,TABLE2|"
-			      "--load-only=table1,table2] FILE1"
+			      "--load-only=TABLE1,TABLE2] FILE1"
 			      " [FILE2] [...]"),
 			      N_("Load host ACPI tables and tables "
 			      "specified by arguments."),
