@@ -23,12 +23,16 @@
 #include <grub/mm.h>
 #include <grub/kernel.h>
 #include <grub/acpi.h>
+#include <grub/loader.h>
 
 void
 grub_halt (void)
 {
-  grub_machine_fini ();
+  grub_machine_fini (GRUB_LOADER_FLAG_NORETURN);
+#if !defined(__ia64__) && !defined(__arm__) && !defined(__aarch64__) && \
+    !defined(__riscv)
   grub_acpi_halt ();
+#endif
   efi_call_4 (grub_efi_system_table->runtime_services->reset_system,
               GRUB_EFI_RESET_SHUTDOWN, GRUB_EFI_SUCCESS, 0, NULL);
 
