@@ -30,8 +30,4 @@ char *grub_resolve_relative_path (const char *base, const char *path);
 
 char *grub_get_dirname (const char *file_path);
 
-int grub_gui_get_named_color (const char *name, grub_gui_color_t *color);
-
-grub_err_t grub_gui_parse_color (const char *s, grub_gui_color_t *color);
-
 #endif /* GRUB_GUI_STRING_UTIL_HEADER */
